package ru.dbotthepony.mc.otm.android

import com.google.common.collect.ImmutableList
import net.minecraft.ChatFormatting
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.util.INBTSerializable
import ru.dbotthepony.mc.otm.capability.android.AndroidCapabilityPlayer
import ru.dbotthepony.mc.otm.client.render.SkinElement
import ru.dbotthepony.mc.otm.readNbt
import ru.dbotthepony.mc.otm.set
import ru.dbotthepony.mc.otm.writeNbt
import java.io.InputStream
import java.io.OutputStream

abstract class AndroidResearch(val type: AndroidResearchType<*>, val capability: AndroidCapabilityPlayer) : INBTSerializable<CompoundTag> {
	var isResearched = false
		protected set

	/**
	 * Whenever there are changes to network
	 */
	var isDirty = false

	/**
	 * Called when it is required to network everything again
	 */
	abstract fun invalidateNetwork()

	fun unResearch() {
		if (!isResearched) {
			return
		}

		onUnResearch()
		isResearched = false
		isDirty = true
	}

	abstract fun onUnResearch()
	abstract fun onResearched()

	/**
	 * Consumes required resources for researching this technology.
	 *
	 * Returns true if all required resources were consumed, false otherwise.
	 *
	 * This function MUST be atomic, hence it SHOULD NOT consume ANY resources if it returns false.
	 */
	abstract fun consumeResearchCost(simulate: Boolean): Boolean

	/**
	 * Grants all (or some) resources back consumed by [consumeResearchCost].
	 *
	 * Returns true whenever player accepted all resources refunded, false otherwise.
	 */
	abstract fun refund(simulate: Boolean): Boolean

	open fun writeNetwork(buff: OutputStream) {
		buff.writeNbt(serializeNBT())
	}

	open fun readNetwork(buff: InputStream) {
		deserializeNBT(buff.readNbt())
	}

	open val canResearch: Boolean get() {
		if (!consumeResearchCost(simulate = true)) {
			return false
		}

		for (research in prerequisites) {
			if (!capability.getResearch(research).isResearched) {
				return false
			}
		}

		for (research in blockedBy) {
			if (capability.getResearch(research).isResearched) {
				return false
			}
		}

		return true
	}

	fun research(force: Boolean = false): Boolean {
		if (isResearched) {
			return false
		}

		if (force) {
			onResearched()
			isResearched = true
			isDirty = true
			return true
		} else if (consumeResearchCost(false)) {
			onResearched()
			isResearched = true
			isDirty = true
			return true
		}

		return false
	}

	/**
	 * List of all tooltip lines for this research
	 */
	open val tooltipLines: List<Component> by lazy {
		return@lazy listOf(type.displayName.copy().withStyle(ChatFormatting.WHITE))
	}

	/**
	 * List of all tooltip lines for this research, for local player at research screen (clientside only)
	 */
	open val screenTooltipLines: List<Component> by this::tooltipLines

	/**
	 * This should return whatever can be used by other code to display this research's name
	 */
	open val tooltipHeader: Component by lazy { tooltipLines[0] }

	/**
	 * This should return whatever can be used by other code to display this research's name,
	 * for local player at research screen (clientside only)
	 */
	open val screenTooltipHeader: Component by lazy { screenTooltipLines[0] }

	open val skinIcon: SkinElement? get() = null
	open val stackIcon: ItemStack? get() = null
	open val iconText: Component? get() = null

	override fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["researched"] = isResearched
		}
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		isResearched = nbt.getBoolean("researched")
	}

	inline val prerequisites get() = type.prerequisites
	inline val unlocks get() = type.unlocks
	inline val blockedBy get() = type.blockedBy
	inline val blocking get() = type.blocking
}
