package ru.dbotthepony.mc.otm.capability.android

import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.core.Direction
import net.minecraft.world.entity.LivingEntity
import net.minecraftforge.common.capabilities.ICapabilityProvider
import net.minecraftforge.common.util.INBTSerializable
import net.minecraft.nbt.CompoundTag
import net.minecraft.world.item.ItemStack
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidFeature
import java.lang.Runnable
import net.minecraft.server.level.ServerPlayer
import net.minecraftforge.event.entity.living.LivingHurtEvent
import net.minecraft.nbt.ListTag
import net.minecraft.nbt.Tag
import net.minecraft.resources.ResourceLocation
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import net.minecraftforge.network.PacketDistributor
import net.minecraft.world.effect.MobEffect
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import net.minecraftforge.energy.CapabilityEnergy
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent
import net.minecraft.world.effect.MobEffects
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.eventbus.api.SubscribeEvent
import net.minecraftforge.event.entity.living.LivingEvent.LivingUpdateEvent
import net.minecraftforge.eventbus.api.EventPriority
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.extractEnergy
import ru.dbotthepony.mc.otm.capability.receiveEnergy
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.ifHas
import ru.dbotthepony.mc.otm.ifPresentK
import ru.dbotthepony.mc.otm.network.AndroidFeatureRemovePacket
import ru.dbotthepony.mc.otm.network.AndroidFeatureSyncPacket
import ru.dbotthepony.mc.otm.network.AndroidNetworkChannel
import ru.dbotthepony.mc.otm.orNull
import ru.dbotthepony.mc.otm.registry.MRegistry
import ru.dbotthepony.mc.otm.registry.StatNames
import ru.dbotthepony.mc.otm.set
import java.util.*

open class AndroidCapability(final override val entity: LivingEntity) : ICapabilityProvider, IAndroidCapability, INBTSerializable<CompoundTag> {
	protected var battery = ImpreciseFraction.ZERO
	protected var maxBattery = ImpreciseFraction(60000)

	override var batteryItemStack: ItemStack = ItemStack.EMPTY

	protected val features = Object2ObjectArrayMap<AndroidFeatureType<*>, AndroidFeature>()
	protected val networkQueue = ArrayList<Any>()
	protected val queuedTicks = ArrayList<Runnable>()
	protected var tickedOnce = false

	protected fun addFeature(feature: AndroidFeature): Boolean {
		if (features.containsKey(feature.type)) return false
		features[feature.type] = feature

		if (!entity.level.isClientSide) {
			queuedTicks.add(feature::applyModifiers)
		}

		if (entity is ServerPlayer) {
			sendNetwork(AndroidFeatureSyncPacket(feature))
		}

		return true
	}

	@Suppress("unchecked_cast")
	override fun <T : AndroidFeature> addFeature(feature: AndroidFeatureType<T>): T {
		val get = features[feature]
		if (get != null) return get as T

		val factory = feature.create(this)

		features[feature] = factory

		if (!entity.level.isClientSide) {
			queuedTicks.add(factory::applyModifiers)
		}

		if (entity is ServerPlayer) {
			sendNetwork(AndroidFeatureSyncPacket(factory))
		}

		return factory
	}

	override fun removeFeature(feature: AndroidFeatureType<*>): Boolean {
		val removed = features.remove(feature)

		if (removed != null) {
			if (!entity.level.isClientSide) {
				queuedTicks.add {
					removed.removeModifiers()
				}
			}

			if (entity is ServerPlayer) {
				sendNetwork(AndroidFeatureRemovePacket(removed.type))
			}
		}

		return removed != null
	}

	override fun hasFeature(feature: AndroidFeatureType<*>): Boolean {
		return features.containsKey(feature)
	}

	override fun hasFeatureLevel(feature: AndroidFeatureType<*>, level: Int): Boolean {
		val get = features[feature] ?: return false
		return get.level >= level
	}

	@Suppress("unchecked_cast")
	override fun <T : AndroidFeature> getFeature(feature: AndroidFeatureType<T>): T? {
		return features[feature] as T?
	}

	override fun onHurt(event: LivingHurtEvent) {
		for (feature in features.values) {
			feature.onHurt(event)
		}
	}

	override fun serializeNBT(): CompoundTag {
		val tag = CompoundTag()

		tag["energy_stored"] = battery.serializeNBT()
		tag["energy_stored_max"] = maxBattery.serializeNBT()
		tag["battery"] = batteryItemStack.serializeNBT()

		val featureList = ListTag()

		for (feature in features.values) {
			val featureNbt = feature.serializeNBT()

			featureNbt["id"] = feature.type.registryName!!.toString()
			featureList.add(featureNbt)
		}

		tag["features"] = featureList
		return tag
	}

	override fun deserializeNBT(compound: CompoundTag?) {
		compound!!

		compound.ifHas("energy_stored") {
			battery = ImpreciseFraction.deserializeNBT(it)
		}

		compound.ifHas("energy_stored_max") {
			maxBattery = ImpreciseFraction.deserializeNBT(it)
		}

		compound.ifHas("battery", CompoundTag::class.java) {
			batteryItemStack = ItemStack.of(it)
		}

		features.clear()

		val featuresNbt = compound.getList("features", Tag.TAG_COMPOUND.toInt())

		for (tag in featuresNbt) {
			if (tag is CompoundTag) {
				val feature = MRegistry.ANDROID_FEATURES.getValue(ResourceLocation(tag.getString("id")))

				if (feature?.isApplicable(this) == true) {
					val instance = feature.create(this)

					instance.deserializeNBT(tag)
					addFeature(instance)

					if (!entity.level.isClientSide) {
						queuedTicks.add(instance::applyModifiers)
					}
				}
			}
		}
	}

	fun dropBattery() {
		if (batteryItemStack.isEmpty) return
		entity.spawnAtLocation(batteryItemStack)
		batteryItemStack = ItemStack.EMPTY
	}

	protected fun sendNetwork(packet: Any) {
		if (entity is ServerPlayer) {
			if (tickedOnce) {
				AndroidNetworkChannel.send(entity, packet)
			} else {
				networkQueue.add(packet)
			}
		}
	}

	protected open fun tickInnerClient() {}
	protected open fun tickInnerClientAlways() {}

	override fun tickClient() {
		queuedTicks.clear()

		if (!entity.isAlive)
			return

		tickInnerClientAlways()

		if (isAndroid) {
			tickInnerClient()

			for (feature in features.values) {
				feature.tickClient()
			}
		}
	}

	override fun tick() {
		if (!entity.isAlive) return

		tickServerAlways()

		if (isAndroid) {
			tickServer()

			for (feature in features.values) {
				feature.tickServer()
			}
		}

		for (runnable in queuedTicks) {
			runnable.run()
		}

		queuedTicks.clear()
	}

	protected open fun tickServerAlways() {}

	protected open fun tickServer() {
		if (entity.airSupply < entity.maxAirSupply)
			entity.airSupply = entity.maxAirSupply

		for (effect in UNAFFECTED_EFFECTS)
			if (entity.hasEffect(effect))
				entity.removeEffect(effect)

		if (!batteryItemStack.isEmpty && battery < maxBattery) {
			batteryItemStack.getCapability(CapabilityEnergy.ENERGY).ifPresent {
				if (it is IMatteryEnergyStorage) {
					battery += it.extractEnergyInner(maxBattery - battery, false)
				} else {
					battery += it.extractEnergy(maxBattery - battery, false)
				}
			}
		}
	}

	override fun extractEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return ImpreciseFraction.ZERO
	}

	override fun extractEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		@Suppress("name_shadowing")
		var howMuch = howMuch
		var drained = ImpreciseFraction.ZERO

		if (!batteryItemStack.isEmpty) {
			batteryItemStack.getCapability(CapabilityEnergy.ENERGY).ifPresent {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.extractEnergyOuter(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				} else {
					val extracted = it.extractEnergy(howMuch, simulate)
					drained += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero) {
				if (!simulate && entity is ServerPlayer) {
					entity.awardStat(StatNames.POWER_CONSUMED, drained.toInt() * 10)
				}

				return drained
			}
		}

		val new = (battery - howMuch).moreThanZero()
		drained += battery - new

		if (!simulate) {
			battery = new

			if (entity is ServerPlayer) {
				entity.awardStat(StatNames.POWER_CONSUMED, drained.toInt() * 10)
			}
		}

		return drained
	}

	override fun receiveEnergyOuter(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		var howMuch = howMuch
		var received = ImpreciseFraction.ZERO

		if (!batteryItemStack.isEmpty) {
			batteryItemStack.getCapability(CapabilityEnergy.ENERGY).ifPresent {
				if (it is IMatteryEnergyStorage) {
					val extracted = it.receiveEnergyOuter(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				} else {
					val extracted = it.receiveEnergy(howMuch, simulate)
					received += extracted
					howMuch -= extracted
				}
			}

			if (howMuch.isZero) {
				return received
			}
		}

		val new = (battery + howMuch).coerceAtMost(maxBattery)
		received += new - battery

		if (!simulate) {
			battery = new
		}

		return received
	}

	override fun receiveEnergyInner(howMuch: ImpreciseFraction, simulate: Boolean): ImpreciseFraction {
		return receiveEnergyOuter(howMuch, simulate)
	}

	override var batteryLevel: ImpreciseFraction
		get() {
			if (!batteryItemStack.isEmpty) {
				val resolver = batteryItemStack.getCapability(CapabilityEnergy.ENERGY).resolve()

				if (resolver.isPresent) {
					val it = resolver.get()

					if (it is IMatteryEnergyStorage) {
						return battery + it.batteryLevel
					} else {
						return battery + it.energyStored
					}
				}
			}

			return battery
		}
		set(value) {
			battery = value
		}

	override var maxBatteryLevel: ImpreciseFraction
		get() {
			if (batteryItemStack != ItemStack.EMPTY) {
				val resolver = batteryItemStack.getCapability(CapabilityEnergy.ENERGY).resolve()

				if (resolver.isPresent) {
					val it = resolver.get()

					if (it is IMatteryEnergyStorage) {
						return maxBattery + it.maxBatteryLevel
					} else {
						return maxBattery + it.maxEnergyStored
					}
				}
			}

			return maxBattery
		}
		set(value) {
			maxBattery = value
		}

	private val resolver = LazyOptional.of { this }

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		return if (cap === MatteryCapability.ANDROID) {
			resolver.cast()
		} else LazyOptional.empty()
	}

	@Suppress("unused")
	companion object {
		@JvmField
		val UNAFFECTED_EFFECTS = ObjectArraySet<MobEffect>()

		@JvmStatic
		fun registerEffects(event: FMLCommonSetupEvent) {
			UNAFFECTED_EFFECTS.add(MobEffects.CONDUIT_POWER)
			UNAFFECTED_EFFECTS.add(MobEffects.HEAL)
			// maybe it makes things go haywire idk
			// UNAFFECTED_EFFECTS.add(MobEffects.HARM);
			UNAFFECTED_EFFECTS.add(MobEffects.REGENERATION)
			UNAFFECTED_EFFECTS.add(MobEffects.WATER_BREATHING)
			UNAFFECTED_EFFECTS.add(MobEffects.POISON)
			// even skeletons can be withered
			// UNAFFECTED_EFFECTS.add(MobEffects.WITHER);
			UNAFFECTED_EFFECTS.add(MobEffects.HEALTH_BOOST)
			UNAFFECTED_EFFECTS.add(MobEffects.ABSORPTION)
			UNAFFECTED_EFFECTS.add(MobEffects.SATURATION)
			UNAFFECTED_EFFECTS.add(MobEffects.DOLPHINS_GRACE)
		}

		@SubscribeEvent
		fun onLivingTick(event: LivingUpdateEvent) {
			val ent = event.entity

			if (ent.level.isClientSide) {
				ent.getCapability(MatteryCapability.ANDROID).ifPresentK {
					it.tickClient()
				}
			} else {
				ent.getCapability(MatteryCapability.ANDROID).ifPresentK {
					it.tick()
				}
			}
		}

		@SubscribeEvent(priority = EventPriority.LOWEST)
		fun onHurtEvent(event: LivingHurtEvent) {
			if (event.isCanceled) return

			event.entity.getCapability(MatteryCapability.ANDROID).ifPresentK { cap: IAndroidCapability -> cap.onHurt(event) }
		}
	}
}

val ICapabilityProvider.android get() = getCapability(MatteryCapability.ANDROID).orNull()
