package ru.dbotthepony.mc.otm.menu

import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.*
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.network.PacketDistributor
import ru.dbotthepony.mc.otm.container.ItemFilter
import ru.dbotthepony.mc.otm.container.ItemFilterNetworkSlot
import ru.dbotthepony.mc.otm.menu.data.MultiByteDataContainer
import ru.dbotthepony.mc.otm.menu.widget.AbstractWidget
import ru.dbotthepony.mc.otm.network.MatteryNetworking
import java.util.Collections

abstract class MatteryMenu @JvmOverloads protected constructor(
	p_38851_: MenuType<*>?,
	p_38852_: Int,
	val inventory: Inventory,
	val tile: BlockEntity? = null
) : AbstractContainerMenu(p_38851_, p_38852_) {
	val ply: Player get() = inventory.player

	private val _matteryWidgets = ArrayList<AbstractWidget>()
	val matteryWidgets: List<AbstractWidget> = Collections.unmodifiableList(_matteryWidgets)

	private val _playerInventorySlots = ArrayList<MatterySlot>()
	val playerInventorySlots: List<MatterySlot> = Collections.unmodifiableList(_playerInventorySlots)
	val multiByteContainers = ArrayList<MultiByteDataContainer>()

	protected val lockedInventorySlots: MutableSet<Int> = HashSet()
	protected open fun isInventorySlotLocked(index: Int): Boolean = lockedInventorySlots.contains(index)

	private val _filterSlots = ArrayList<ItemFilterNetworkSlot>()
	val filterSlots: List<ItemFilterNetworkSlot> = Collections.unmodifiableList(_filterSlots)

	fun addFilterSlots(slots: ItemFilter): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(slots.size)

		for (i in 0 until slots.size) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, slots).also(result::add))
		}

		return result
	}

	fun addFilterSlots(amount: Int): List<ItemFilterNetworkSlot> {
		val result = ArrayList<ItemFilterNetworkSlot>(amount)

		for (i in 0 until amount) {
			_filterSlots.add(ItemFilterNetworkSlot(i, _filterSlots.size, null).also(result::add))
		}

		return result
	}

	fun addWidget(widget: AbstractWidget): Int {
		val indexOf = _matteryWidgets.indexOf(widget)

		if (indexOf == -1) {
			_matteryWidgets.add(widget)
			return _matteryWidgets.size - 1
		}

		return indexOf
	}

	fun getWidget(index: Int): AbstractWidget? {
		if (index !in _matteryWidgets.indices) {
			return null
		}

		return _matteryWidgets[index]
	}

	fun addDataContainer(container: MultiByteDataContainer) {
		if (multiByteContainers.contains(container))
			throw IllegalStateException("Already containing $container")

		multiByteContainers.add(container)
		container.slotID = multiByteContainers.size - 1
	}

	fun getDataContainer(index: Int): MultiByteDataContainer? {
		if (index !in 0 until multiByteContainers.size) {
			return null
		}

		return multiByteContainers[index]
	}

	protected var inventorySlotIndexStart = 0
	protected var inventorySlotIndexEnd = 0

	protected fun addInventorySlots() {
		var first = true

		for (i in 0..2) {
			for (j in 0..8) {
				val slot: MatterySlot = object : MatterySlot(inventory, j + i * 9 + 9, 8 + j * 18, 14 + i * 18) {
					override fun mayPlace(p_40231_: ItemStack): Boolean {
						return !isInventorySlotLocked(index)
					}

					override fun mayPickup(p_40228_: Player): Boolean {
						return !isInventorySlotLocked(index)
					}
				}

				_playerInventorySlots.add(slot)
				addSlot(slot)

				if (first) {
					first = false
					inventorySlotIndexStart = slot.index
				}
			}
		}

		var last: MatterySlot? = null

		for (k in 0..8) {
			last = object : MatterySlot(inventory, k, 8 + k * 18, 14 + 58) {
				override fun mayPlace(p_40231_: ItemStack): Boolean {
					return !isInventorySlotLocked(index)
				}

				override fun mayPickup(p_40228_: Player): Boolean {
					return !isInventorySlotLocked(index)
				}
			}

			addSlot(last)
			_playerInventorySlots.add(last)
		}

		inventorySlotIndexEnd = last!!.index
	}

	private val playerPacketDistributor = PacketDistributor.PLAYER.with { ply as ServerPlayer }

	override fun broadcastChanges() {
		for (widget in _matteryWidgets) {
			widget.updateServer()
		}

		for (data in multiByteContainers) {
			if (data.dirty) {
				data.dirty = false
				MatteryNetworking.CHANNEL.send(playerPacketDistributor, data.makePacket())
			}
		}

		super.broadcastChanges()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges()

			if (packet != null) {
				MatteryNetworking.CHANNEL.send(consumer, packet)
			}
		}
	}

	override fun broadcastFullState() {
		for (widget in _matteryWidgets) {
			widget.updateServer()
		}

		for (data in multiByteContainers) {
			MatteryNetworking.CHANNEL.send(playerPacketDistributor, data.makePacket())
		}

		super.broadcastFullState()

		val consumer = PacketDistributor.PLAYER.with { ply as ServerPlayer }

		for (slot in _filterSlots) {
			val packet = slot.sendChanges(true)

			if (packet != null) {
				MatteryNetworking.CHANNEL.send(consumer, packet)
			}
		}
	}

	override fun stillValid(player: Player): Boolean {
		if (tile == null) return true

		if (player.level.getBlockEntity(tile.blockPos) !== tile) {
			return false
		}

		val pos = tile.blockPos

		return player.distanceToSqr(pos.x.toDouble() + 0.5, pos.y.toDouble() + 0.5, pos.z.toDouble() + 0.5) <= 64.0
	}

	abstract val storageSlots: Collection<Slot>

	// This method receive Player interactor and slot_index where Shift + Right click occurred
	// It shall return item stack that got moved
	override fun quickMoveStack(ply: Player, slotIndex: Int): ItemStack {
		if (storageSlots.isEmpty()) {
			return ItemStack.EMPTY
		}

		var moveToPlayer: Boolean? = null

		for (slot in storageSlots) {
			if (slot.index == slotIndex) {
				moveToPlayer = true
				break
			}
		}

		if (moveToPlayer == null) {
			for (slot in playerInventorySlots) {
				if (slot.index == slotIndex) {
					moveToPlayer = false
					break
				}
			}
		}

		if (moveToPlayer == null) {
			return ItemStack.EMPTY
		}

		var moved = ItemStack.EMPTY
		val initialSlot = slots[slotIndex]

		if (initialSlot.hasItem()) {
			if (!initialSlot.mayPickup(ply)) {
				return moved
			}

			val initialItem = initialSlot.item
			moved = initialItem.copy()

			if (moveToPlayer) {
				if (!moveItemStackTo(initialItem, playerInventorySlots)) {
					return ItemStack.EMPTY
				}
			} else {
				if (!moveItemStackTo(initialItem, storageSlots)) {
					return ItemStack.EMPTY
				}
			}

			if (initialItem.isEmpty) {
				initialSlot.set(ItemStack.EMPTY)
			} else {
				initialSlot.setChanged()
			}
		}

		return moved
	}

	fun quickMoveToInventory(stack: ItemStack, simulate: Boolean): ItemStack {
		if (inventorySlotIndexStart == 0 && inventorySlotIndexEnd == 0) {
			return stack
		}

		return moveItemStackToSlots(stack, inventorySlotIndexStart, inventorySlotIndexEnd, simulate = simulate)
	}

	override fun moveItemStackTo(
		item: ItemStack,
		initialSlot: Int,
		finalSlot: Int,
		inverse: Boolean
	): Boolean {
		val remainder = moveItemStackToSlots(item, initialSlot, finalSlot, inverse)

		if (remainder.count == item.count) {
			return false
		}

		item.count = remainder.count
		return true
	}

	fun moveItemStackTo(
		item: ItemStack,
		slots: Collection<Slot>
	): Boolean {
		val remainder = moveItemStackToSlots(item, slots)

		if (remainder.count == item.count) {
			return false
		}

		item.count = remainder.count
		return true
	}

	fun moveItemStackToSlots(item: ItemStack, slots: Collection<Slot>, simulate: Boolean = false): ItemStack {
		val copy = item.copy()

		// first pass - stack with existing slots
		if (copy.isStackable) {
			for (slot in slots) {
				val limit = slot.getMaxStackSize(copy)

				if (limit > slot.item.count && ItemStack.isSameItemSameTags(slot.item, copy) && slot.mayPlace(item)) {
					val newCount = (slot.item.count + copy.count).coerceAtMost(limit)
					val diff = newCount - slot.item.count
					copy.count -= diff

					if (!simulate) {
						slot.item.count += diff
						slot.setChanged()
					}

					if (copy.isEmpty) {
						return copy
					}
				}
			}
		}

		// second pass - drop stack into first free slot
		for (slot in slots) {
			val limit = slot.getMaxStackSize(copy)

			if (!slot.hasItem() && slot.mayPlace(item)) {
				val newCount = copy.count.coerceAtMost(limit)

				if (!simulate) {
					slot.set(copy.copy().also { it.count = newCount })
					slot.setChanged()
				}

				copy.count -= newCount

				if (copy.isEmpty) {
					return copy
				}
			}
		}

		return copy
	}

	fun moveItemStackToSlots(item: ItemStack, initialSlot: Int, finalSlot: Int, inverse: Boolean = false, simulate: Boolean = false): ItemStack {
		if (initialSlot > finalSlot) {
			return item
		}

		require(initialSlot >= 0) { "Invalid initial slot $initialSlot" }
		require(finalSlot < slots.size) { "Final slot $finalSlot is bigger than total size of array of ${slots.size}" }

		val slots = ArrayList<Slot>(finalSlot - initialSlot + 1)

		for (i in (if (inverse) finalSlot downTo initialSlot else initialSlot .. finalSlot)) {
			slots.add(slots[i])
		}

		return moveItemStackToSlots(item, slots, simulate)
	}

	public override fun addDataSlots(p_38885_: ContainerData) {
		super.addDataSlots(p_38885_)
	}
}
