package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.world.item.ItemStack
import net.minecraftforge.network.NetworkDirection.PLAY_TO_CLIENT
import net.minecraftforge.network.NetworkDirection.PLAY_TO_SERVER
import net.minecraftforge.network.NetworkEvent
import ru.dbotthepony.mc.otm.android.AndroidFeature
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidResearch
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.capability.android.*
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.readImpreciseFraction
import ru.dbotthepony.mc.otm.core.writeImpreciseFraction
import ru.dbotthepony.mc.otm.menu.AndroidStationMenu
import ru.dbotthepony.mc.otm.registry.MRegistry
import java.io.ByteArrayInputStream
import java.io.OutputStream
import java.util.function.Supplier

sealed class AndroidStatusPacket : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {}

	protected abstract fun play(capability: AndroidCapabilityPlayer)

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			play(minecraft.player?.android as? AndroidCapabilityPlayer ?: return@enqueueWork)
		}
	}
}

object IsAndroidPacket : AndroidStatusPacket() {
	override fun play(capability: AndroidCapabilityPlayer) {
		capability.isAndroid = true
	}
}

object IsNotAndroidPacket : AndroidStatusPacket() {
	override fun play(capability: AndroidCapabilityPlayer) {
		capability.isAndroid = false
	}
}

fun IsAndroidPacket(status: Boolean) = if (status) IsAndroidPacket else IsNotAndroidPacket

object WillBecomeAndroidPacket : AndroidStatusPacket() {
	override fun play(capability: AndroidCapabilityPlayer) {
		capability.willBecomeAndroid = true
	}
}

object WillNotBecomeAndroidPacket : AndroidStatusPacket() {
	override fun play(capability: AndroidCapabilityPlayer) {
		capability.willBecomeAndroid = false
	}
}

fun WillBecomeAndroidPacket(status: Boolean) = if (status) WillBecomeAndroidPacket else WillNotBecomeAndroidPacket

class AndroidResearchRequestPacket(val type: AndroidResearchType<*>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(MRegistry.ANDROID_RESEARCH.getID(type))
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val ply = context.get().sender ?: return@enqueueWork
			val android = ply.android as? AndroidCapabilityPlayer ?: return@enqueueWork

			if (!android.isAndroid || ply.containerMenu !is AndroidStationMenu)
				return@enqueueWork

			android.getResearch(type).research()
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidResearchRequestPacket {
			return AndroidResearchRequestPacket(MRegistry.ANDROID_RESEARCH.getValue(buff.readInt()))
		}
	}
}

private fun getData(invoke: (stream: OutputStream) -> Unit): FastByteArrayOutputStream {
	val stream = FastByteArrayOutputStream()
	invoke(stream)
	return stream
}

class AndroidResearchSyncPacket(val type: AndroidResearchType<*>, val dataList: FastByteArrayOutputStream?, val dataBytes: ByteArray?) : MatteryPacket {
	constructor(feature: AndroidResearch) : this(feature.type, getData(feature::writeNetwork), null)

	override fun write(buff: FriendlyByteBuf) {
		dataList ?: throw NullPointerException("No byte list is present")
		buff.writeInt(MRegistry.ANDROID_RESEARCH.getID(type))
		buff.writeBytes(dataList.array, 0, dataList.length)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		dataBytes ?: throw NullPointerException("No data bytes array is present")

		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.android as? AndroidCapabilityPlayer ?: return@enqueueWork

			android.getResearch(type).readNetwork(ByteArrayInputStream(dataBytes))
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidResearchSyncPacket {
			return AndroidResearchSyncPacket(
				MRegistry.ANDROID_RESEARCH.getValue(buff.readInt()),
				null, ByteArray(buff.readableBytes()).also { buff.readBytes(it) }
			)
		}
	}
}

class AndroidFeatureSyncPacket(val type: AndroidFeatureType<*>, val dataList: FastByteArrayOutputStream?, val dataBytes: ByteArray?) : MatteryPacket {
	constructor(feature: AndroidFeature) : this(feature.type, getData(feature::writeNetwork), null)

	override fun write(buff: FriendlyByteBuf) {
		dataList ?: throw NullPointerException("No byte list is present")
		buff.writeInt(MRegistry.ANDROID_FEATURES.getID(type))
		buff.writeBytes(dataList.array, 0, dataList.length)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		dataBytes ?: throw NullPointerException("No data bytes array is present")

		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.android as? AndroidCapabilityPlayer ?: return@enqueueWork

			android.computeIfAbsent(type).readNetwork(ByteArrayInputStream(dataBytes))
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidFeatureSyncPacket {
			return AndroidFeatureSyncPacket(
				MRegistry.ANDROID_FEATURES.getValue(buff.readInt()),
				null, ByteArray(buff.readableBytes()).also { buff.readBytes(it) }
			)
		}
	}
}

class AndroidFeatureRemovePacket(val type: AndroidFeatureType<*>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(MRegistry.ANDROID_FEATURES.getID(type))
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.android as? AndroidCapabilityPlayer ?: return@enqueueWork

			android.removeFeature(type)
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidFeatureRemovePacket {
			return AndroidFeatureRemovePacket(MRegistry.ANDROID_FEATURES.getValue(buff.readInt()))
		}
	}
}

class AndroidBatteryItemPacket(val item: ItemStack) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeItem(item)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.android as? AndroidCapabilityPlayer ?: return@enqueueWork

			android.batteryItemStack = item
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidBatteryItemPacket {
			return AndroidBatteryItemPacket(buff.readItem())
		}
	}
}

class AndroidEnergyLevelPacket(val level: ImpreciseFraction, val isMaxEnergy: Boolean) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeImpreciseFraction(level)
		buff.writeBoolean(isMaxEnergy)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.android as? AndroidCapabilityPlayer ?: return@enqueueWork

			if (isMaxEnergy) {
				android.maxBatteryLevel = level
			} else {
				android.batteryLevel = level
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidEnergyLevelPacket {
			return AndroidEnergyLevelPacket(buff.readImpreciseFraction(), buff.readBoolean())
		}
	}
}

object AndroidNetworkChannel : MatteryNetworkChannel(
	version = "1",
	name = "android"
) {
	fun register() {
		add(IsAndroidPacket::class.java, { IsAndroidPacket }, PLAY_TO_CLIENT)
		add(IsNotAndroidPacket::class.java, { IsNotAndroidPacket }, PLAY_TO_CLIENT)
		add(WillBecomeAndroidPacket::class.java, { WillBecomeAndroidPacket }, PLAY_TO_CLIENT)
		add(WillNotBecomeAndroidPacket::class.java, { WillNotBecomeAndroidPacket }, PLAY_TO_CLIENT)

		add(AndroidResearchRequestPacket::class.java, AndroidResearchRequestPacket.Companion::read, PLAY_TO_SERVER)
		add(AndroidResearchSyncPacket::class.java, AndroidResearchSyncPacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidFeatureSyncPacket::class.java, AndroidFeatureSyncPacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidFeatureRemovePacket::class.java, AndroidFeatureRemovePacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidBatteryItemPacket::class.java, AndroidBatteryItemPacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidEnergyLevelPacket::class.java, AndroidEnergyLevelPacket.Companion::read, PLAY_TO_CLIENT)
	}
}
